//
// soft synth test
//
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <ctype.h>
#include <windows.h>

//lib
#pragma comment(lib,"user32")
#pragma comment(lib,"winmm")

//def
#define SND_PI             (3.141592653589793f)
#define SND_NOTE_DT        (1.059463100070972f)
#define SND_NOTE_BASE      (7.75f / (SND_TYPE)SND_SAMPLES)
#define SND_BITS           16
#define SND_SAMPLES        44100
#define SND_SAMPLE_DIV     10
#define SND_SAMPLE_UNIT    (SND_SAMPLES / SND_SAMPLE_DIV)
#define SND_SAMPLES_DT     (1.0f / (float)SND_SAMPLES);
#define SND_CH             1
#define SND_TRACK_MAX      4
#define SND_SEC            60 * 3
#define SND_LENGTH         (SND_CH * SND_SAMPLES * SND_SEC)
#define SND_UNIT_NUM       ( (SND_LENGTH / SND_CH) / SND_SAMPLE_UNIT )
#define SND_TYPE           float
#define SND_CODE_END       0xFF
#define SND_CODE_CUT       0x7F

//global data
static HWAVEOUT hWOut = NULL;
static WAVEFORMATEX wf;
static WAVEHDR wh;
static SND_TYPE canvas[SND_LENGTH];
static SND_TYPE raw_buffer[SND_LENGTH];

//sound ch
struct SND_TRACK {
  int   type;
  float  t;
  float dt;
  float gain;
  float dgain;
};
SND_TRACK snd_track[SND_TRACK_MAX];

void snd_set_track(int ch, int type, float gain, float dgain) {
  if(ch < SND_TRACK_MAX) {
    SND_TRACK *w = &(snd_track[ch]);
    w->type = type;
    w->t  = 0;
    w->dt = 0;
    w->gain = gain;
    w->dgain = dgain;
  }
}

float snd_rand() {
  return (float)(rand() / 32767.0f);
}

float snd_round(float a) {
  return (float)(int)(a + 0.5f);
}

void snd_term() {
  if(hWOut) {
    waveOutClose(hWOut);
    hWOut = NULL;
  }
  memset(canvas, 0, sizeof(canvas));
  memset(raw_buffer, 0, sizeof(raw_buffer));
  memset(snd_track, 0, sizeof(snd_track));
}

void snd_reset() {
  snd_term();
}


void snd_mix(float gain) {
  for(int i = 0 ; i < SND_LENGTH ; i++) {
    raw_buffer[i] += (canvas[i] * gain);
  }
}

void snd_low_pass(float k) {
  for(int i = 1 ; i < SND_LENGTH ; i++) {
    canvas[i] = canvas[i] + ((canvas[i - 1] - canvas[i]) * k);
  }
}

void snd_high_pass(float k) {
  for(int i = 1 ; i < SND_LENGTH ; i++) {
    canvas[i] = canvas[i] - canvas[i - 1] * k;
  }
}

void snd_delay(int t, float k) {
  for(int i = t ; i < SND_LENGTH ; i++) {
    canvas[i] = canvas[i] + (canvas[i - t] * k);
  }
}

void snd_convert() {
  if(SND_BITS != 16) {
    MessageBox(NULL, "Error", "Not support sound bits", MB_OK);
    return;
  }
  short *buf = (short *)raw_buffer;
  for(int i = 0 ; i < SND_LENGTH ; i++) {
    buf[i] = (short)(raw_buffer[i] * 32767.0f);
  }
}

float snd_get_note(int note) {
  float dt = SND_NOTE_BASE;
  for(int i = 0 ; i < note; i++) {
    dt *= SND_NOTE_DT;
  }
  
  return dt;
}

void snd_render_track(int ch, unsigned char *data, float mix_vol) {
  if(ch > SND_TRACK_MAX) return;
  printf("Render Track ch[%d]\n", ch);
  SND_TRACK *w = &(snd_track[ch]);
  memset(canvas, 0, sizeof(canvas));
  float *buf = canvas;
  float t  = 0;
  float dt = 0;
  float gain = 0;
  float dgain = 0.9999;
  for(int n = 0 ; n < SND_UNIT_NUM ; n++, buf += (SND_SAMPLE_UNIT * SND_CH) ) {
    int note = data[n];
    if(note == SND_CODE_END) break;
    if(note == SND_CODE_CUT) {
      dt = 0;
      t  = 0;
      continue;
    }
    if(note != 0x00) {
      t  = 0;
      //TEST
      //dt = snd_get_note(note);
      dt = snd_get_note(note - 3 + (n > 256 ? (n > 512 ? ( n > 768 ? 0 :  7) : -5 ) : 0)  );
      gain = 1;
    }
    if(note != 0   ) t = 0;
    for(int i = 0 ; i < SND_SAMPLE_UNIT; i++) {
      t += dt;
      t -= snd_round(t);
      
      float gt = 0;
      if( (n % 64) < 32) {
        gt = t;
      } else {
        gt = t * SND_PI * 2;
      }
      //TEST
      buf[i] = sin(gt) + sin(gt * 0.3);
      
      buf[i] = sin(buf[i] * ( (n/64) + 1) ) * gain;
      gain *= dgain;
    }
  }
  //TEST
  snd_delay(SND_SAMPLES * 3.61, 0.7);
  snd_delay(SND_SAMPLES * 2.61, 0.3);
  snd_low_pass(0.1);
  snd_high_pass(0.5);
  
  snd_mix(mix_vol);
}

//Play Sound
void snd_play() {
  //convert float -> short
  snd_convert();
  
  //play sound
  memset(&wf, 0, sizeof(wf));
  wf.wFormatTag      = WAVE_FORMAT_PCM;
  wf.nChannels       = SND_CH;
  wf.nSamplesPerSec  = SND_SAMPLES;
  wf.wBitsPerSample  = SND_BITS;
  wf.nBlockAlign     = wf.nChannels * wf.wBitsPerSample / 8;
  wf.nAvgBytesPerSec = wf.nSamplesPerSec * wf.nBlockAlign;
  wf.cbSize          = 0;
  waveOutOpen(&hWOut, WAVE_MAPPER, &wf, 0, 0, CALLBACK_NULL);
  
  memset(&wh, 0, sizeof(wh));
  wh.lpData          = (LPSTR)raw_buffer;
  wh.dwBufferLength  = SND_LENGTH * (wf.wBitsPerSample / 8);
  wh.dwFlags         = 0;
  wh.dwLoops         = 1;
  wh.dwBytesRecorded = 0;
  wh.dwUser          = 0;
  wh.lpNext          = NULL;
  wh.reserved        = 0;
  waveOutPrepareHeader(hWOut , &wh , sizeof(WAVEHDR));
  waveOutWrite(hWOut , &wh , sizeof(WAVEHDR));
}


//ep
int main() {
  snd_reset();
  static unsigned char data1[] = {
    //0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x42, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    0x42, 0x00, 0x00, 0x00,  0x3F, 0x00, 0x41, 0x00,  0x3D, 0x00, 0x38, 0x00,  0x44, 0x00, 0x00, 0x00,
    
    SND_CODE_END
  };
  
  printf("SND_UNIT_NUM = %d\n", SND_UNIT_NUM);

  snd_render_track(0, data1, 0.03);
  
  //int index = 0;
  //while(data1[index] != SND_CODE_END) {
  //  if(data1[index]) {
  //    data1[index] -= 12;
  //  }
  //  index++;
  //}
  //snd_render_track(0, data1, 0.05);

  snd_play();
  MessageBox(NULL,"ՂꂢȂ : OKƏI܂","Wave",MB_OK);
  snd_term();
  return 0;
}


